#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/prctl.h>
#include <unistd.h>
#include <X11/Xlib.h>

char *childname, **childparam;
int childsig = SIGTERM;
pid_t childpid = -1;
int verbose=0;
const float version=0.01;
const char *keys="hvVk:";
#define USAGE "[-hvV] [-k NUMBER] prorgam [program_options...]"
#define HELP "XRun is an X client that runs an arbitrary program, \n\
and kill it when a connection to X server is lost."
#define OPTIONS "\
  -k NUMBER\tkill program with signal NUMBER \n\
  -v\t\tincrease verbosity \n\
  -h\t\tdisplay help message \n\
  -V\t\tdistplay version"
/* TODO help suitable for help2man */

int (*old_error_handler) (Display *);
Display *dpy;

#define debug(...) ((verbose > 0 ) && (error(0, 0, __VA_ARGS__),0))
#define usage(s, e, u, ...) error(s, e, u "\nUsage: " USAGE, ##__VA_ARGS__)

static int error_handler(Display *display)
{
    debug("X server I/O error, killing %d with %d", childpid, childsig);
    if(kill(childpid, childsig))
      error(0, errno, "Cannot kill child '%s' (%d) with %d signal", 
	  childname, childpid, childsig);
    return 0;
}

void child_terminated(int signo)
{
    int status;

    error(0, 0, "Child '%s' (%d) died", childname, childpid);
    wait(&status);
    exit(status);
}

static int wait_forever(void)
{
    XEvent event;
    XNextEvent(dpy, &event); /* This shoud not happen */
    error(0, 0, "Unexpected event from window %lx: %d", ((XAnyEvent *)&event)->window, event.type);
    return event.type;
}

int main(int argc, char *argv[])
{
    int opt, ret=0;

    while ((opt = getopt(argc, argv, keys)) != -1) {
      switch (opt) {
      case 'V': printf("%s %3.2f\n", program_invocation_short_name, version); return 0;
      case 'v': verbose++; break;
      case 'k': childsig = atoi(optarg); break;
      default : error(3, 0, USAGE); break;
      case 'h': printf("Usage: %s %s\n\n", program_invocation_short_name, USAGE);
		printf("%s\n\n",HELP);
		printf("Options:\n%s\n",OPTIONS);
		return 0;
      }
    }

    if(optind>=argc)
	usage(3, 0, "No program specified");

    childname = argv[optind];
    childparam = argv + optind;

    dpy = XOpenDisplay(NULL);
    if (!dpy) 
        usage(5, 0, "Unable to open display %s, NOT executing '%s'",
        	XDisplayName(NULL), childname);

    old_error_handler = XSetIOErrorHandler(error_handler);

    signal(SIGCHLD, child_terminated);
    switch (childpid = fork()) {
    case -1:
	XCloseDisplay(dpy);
        error(1, errno, "Error forking");
    case 0:
	prctl(PR_SET_PDEATHSIG, childsig);
	execvp(childname, childparam);
	error(2, errno, "Error executing '%s'", childname);
    }
    debug("Child (%d) started: %s", childpid, childname);

    /* TODO: change process name */
    /* Non-eager todo:
     * open unmapped window and set resource names
     * for child can also be killed by xkill'ing parent
     */
    return wait_forever();
}
